package ru.dbotthepony.mc.otm.client.render

import com.mojang.blaze3d.vertex.PoseStack
import com.mojang.blaze3d.vertex.Tesselator
import com.mojang.blaze3d.vertex.VertexConsumer
import com.mojang.math.Matrix4f
import com.mojang.math.Vector3f
import net.minecraft.client.gui.Font
import net.minecraft.core.Vec3i
import net.minecraft.network.chat.Component
import net.minecraft.util.FormattedCharSequence
import ru.dbotthepony.mc.otm.core.*

val tesselator: Tesselator get() = Tesselator.getInstance()

fun VertexConsumer.normal(vector: Vector): VertexConsumer = normal(vector.x.toFloat(), vector.y.toFloat(), vector.z.toFloat())
fun VertexConsumer.vertex(matrix4f: Matrix4f, vector: Vector): VertexConsumer = vertex(matrix4f, vector.x.toFloat(), vector.y.toFloat(), vector.z.toFloat())
fun VertexConsumer.color(color: RGBAColor): VertexConsumer = color(color.red, color.green, color.blue, color.alpha)

fun PoseStack.translate(vector: Vector) = translate(vector.x, vector.y, vector.z)
fun PoseStack.translate(vector: Vec3i) = translate(vector.x.toDouble(), vector.y.toDouble(), vector.z.toDouble())
fun PoseStack.translate(vector: Vector3f) = last().pose().multiplyWithTranslation(vector.x(), vector.y(), vector.z())

fun PoseStack.rotateAroundPoint(point: Vector, axis: Vector, rotation: Float, isDegrees: Boolean = false) {
	val last = last()
	last.pose().rotateAroundPoint(point, axis, rotation, isDegrees)
	last.normal().mul(axis.rotateAroundThis(rotation, isDegrees))
}

fun PoseStack.rotateAroundPoint(point: Vector, rotation: IAngle) {
	val last = last()
	last.pose().rotateAroundPoint(point, rotation)
	// last.normal().mul(rotation.forward().rotateAroundThis(rotation))
}

fun PoseStack.rotateAroundPoint(point: Vector3f, axis: Vector, rotation: Float, isDegrees: Boolean = false) {
	val last = last()
	last.pose().rotateAroundPoint(point, axis, rotation, isDegrees)
	last.normal().mul(axis.rotateAroundThis(rotation, isDegrees))
}

fun PoseStack.rotateAroundPoint(point: Vector3f, rotation: IAngle) {
	val last = last()
	last.pose().rotateAroundPoint(point, rotation)
	// last.normal().mul(rotation.forward().rotateAroundThis(rotation))
}

fun PoseStack.translation(): Vector3f {
	return last().pose().getTranslation()
}

enum class TextAlign {
	TOP_LEFT,
	TOP_CENTER,
	TOP_RIGHT,

	CENTER_LEFT,
	CENTER_CENTER,
	CENTER_RIGHT,

	BOTTOM_LEFT,
	BOTTOM_CENTER,
	BOTTOM_RIGHT,
}

fun Font.drawAligned(poseStack: PoseStack, text: String, align: TextAlign, x: Float, y: Float, color: Int): Int {
	return when (align) {
		TextAlign.TOP_LEFT -> draw(poseStack, text, x, y, color)
		TextAlign.TOP_CENTER -> draw(poseStack, text, x - width(text) / 2f, y, color)
		TextAlign.TOP_RIGHT -> draw(poseStack, text, x - width(text), y, color)

		TextAlign.CENTER_LEFT -> draw(poseStack, text, x, y - lineHeight / 2f, color)
		TextAlign.CENTER_CENTER -> draw(poseStack, text, x - width(text) / 2f, y - lineHeight / 2f, color)
		TextAlign.CENTER_RIGHT -> draw(poseStack, text, x - width(text), y - lineHeight / 2f, color)

		TextAlign.BOTTOM_LEFT -> draw(poseStack, text, x, y - lineHeight, color)
		TextAlign.BOTTOM_CENTER -> draw(poseStack, text, x - width(text) / 2f, y - lineHeight, color)
		TextAlign.BOTTOM_RIGHT -> draw(poseStack, text, x - width(text), y - lineHeight, color)
	}
}

fun Font.drawAligned(poseStack: PoseStack, text: Component, align: TextAlign, x: Float, y: Float, color: Int): Int {
	return when (align) {
		TextAlign.TOP_LEFT -> draw(poseStack, text, x, y, color)
		TextAlign.TOP_CENTER -> draw(poseStack, text, x - width(text) / 2f, y, color)
		TextAlign.TOP_RIGHT -> draw(poseStack, text, x - width(text), y, color)

		TextAlign.CENTER_LEFT -> draw(poseStack, text, x, y - lineHeight / 2f, color)
		TextAlign.CENTER_CENTER -> draw(poseStack, text, x - width(text) / 2f, y - lineHeight / 2f, color)
		TextAlign.CENTER_RIGHT -> draw(poseStack, text, x - width(text), y - lineHeight / 2f, color)

		TextAlign.BOTTOM_LEFT -> draw(poseStack, text, x, y - lineHeight, color)
		TextAlign.BOTTOM_CENTER -> draw(poseStack, text, x - width(text) / 2f, y - lineHeight, color)
		TextAlign.BOTTOM_RIGHT -> draw(poseStack, text, x - width(text), y - lineHeight, color)
	}
}

fun Font.drawAligned(poseStack: PoseStack, text: FormattedCharSequence, align: TextAlign, x: Float, y: Float, color: Int): Int {
	return when (align) {
		TextAlign.TOP_LEFT -> draw(poseStack, text, x, y, color)
		TextAlign.TOP_CENTER -> draw(poseStack, text, x - width(text) / 2f, y, color)
		TextAlign.TOP_RIGHT -> draw(poseStack, text, x - width(text), y, color)

		TextAlign.CENTER_LEFT -> draw(poseStack, text, x, y - lineHeight / 2f, color)
		TextAlign.CENTER_CENTER -> draw(poseStack, text, x - width(text) / 2f, y - lineHeight / 2f, color)
		TextAlign.CENTER_RIGHT -> draw(poseStack, text, x - width(text), y - lineHeight / 2f, color)

		TextAlign.BOTTOM_LEFT -> draw(poseStack, text, x, y - lineHeight, color)
		TextAlign.BOTTOM_CENTER -> draw(poseStack, text, x - width(text) / 2f, y - lineHeight, color)
		TextAlign.BOTTOM_RIGHT -> draw(poseStack, text, x - width(text), y - lineHeight, color)
	}
}

fun Font.drawAligned(poseStack: PoseStack, text: String, align: TextAlign, x: Float, y: Float, color: RGBAColor): Int {
	return drawAligned(poseStack, text, align, x, y, color.toInt())
}

fun Font.drawAligned(poseStack: PoseStack, text: Component, align: TextAlign, x: Float, y: Float, color: RGBAColor): Int {
	return drawAligned(poseStack, text, align, x, y, color.toInt())
}

fun Font.drawAligned(poseStack: PoseStack, text: FormattedCharSequence, align: TextAlign, x: Float, y: Float, color: RGBAColor): Int {
	return drawAligned(poseStack, text, align, x, y, color.toInt())
}

fun Font.drawAlignedLines(poseStack: PoseStack, text: List<Any?>, align: TextAlign, x: Float, y: Float, color: Int) {
	var totalWidth = 0
	var height = 0

	for (line in text) {
		var accumulate = 0

		when (line) {
			is Component -> {
				accumulate += width(line)
				height += lineHeight
			}

			is String -> {
				accumulate += width(line)
				height += lineHeight
			}

			is FormattedCharSequence -> {
				accumulate += width(line)
				height += lineHeight
			}

			else -> height += lineHeight
		}

		totalWidth = totalWidth.coerceAtLeast(accumulate)
	}

	if (height == 0 || totalWidth == 0) {
		return
	}

	var y = y

	when (align) {
		TextAlign.TOP_LEFT -> {
			for (line in text) {
				when (line) {
					is Component -> draw(poseStack, line, x, y, color)
					is String -> draw(poseStack, line, x, y, color)
					is FormattedCharSequence -> draw(poseStack, line, x, y, color)
				}

				y += lineHeight
			}
		}

		TextAlign.TOP_CENTER -> {
			for (line in text) {
				when (line) {
					is Component -> draw(poseStack, line, x - width(line) / 2f, y, color)
					is String -> draw(poseStack, line, x - width(line) / 2f, y, color)
					is FormattedCharSequence -> draw(poseStack, line, x - width(line) / 2f, y, color)
				}

				y += lineHeight
			}
		}

		TextAlign.TOP_RIGHT -> {
			for (line in text) {
				when (line) {
					is Component -> draw(poseStack, line, x - width(line), y, color)
					is String -> draw(poseStack, line, x - width(line), y, color)
					is FormattedCharSequence -> draw(poseStack, line, x - width(line), y, color)
				}

				y += lineHeight
			}
		}

		TextAlign.CENTER_LEFT -> {
			for (line in text) {
				when (line) {
					is Component -> draw(poseStack, line, x, y - height / 2f, color)
					is String -> draw(poseStack, line, x, y - height / 2f, color)
					is FormattedCharSequence -> draw(poseStack, line, x, y - height / 2f, color)
				}

				y += lineHeight
			}
		}

		TextAlign.CENTER_CENTER -> {
			for (line in text) {
				when (line) {
					is Component -> draw(poseStack, line, x - width(line) / 2f, y - height / 2f, color)
					is String -> draw(poseStack, line, x - width(line) / 2f, y - height / 2f, color)
					is FormattedCharSequence -> draw(poseStack, line, x - width(line) / 2f, y - height / 2f, color)
				}

				y += lineHeight
			}
		}

		TextAlign.CENTER_RIGHT -> {
			for (line in text) {
				when (line) {
					is Component -> draw(poseStack, line, x - width(line), y - height / 2f, color)
					is String -> draw(poseStack, line, x - width(line), y - height / 2f, color)
					is FormattedCharSequence -> draw(poseStack, line, x - width(line), y - height / 2f, color)
				}

				y += lineHeight
			}
		}

		TextAlign.BOTTOM_LEFT -> {
			for (line in text) {
				when (line) {
					is Component -> draw(poseStack, line, x, y - height, color)
					is String -> draw(poseStack, line, x, y - height, color)
					is FormattedCharSequence -> draw(poseStack, line, x, y - height, color)
				}

				y += lineHeight
			}
		}

		TextAlign.BOTTOM_CENTER -> {
			for (line in text) {
				when (line) {
					is Component -> draw(poseStack, line, x - width(line) / 2f, y - height, color)
					is String -> draw(poseStack, line, x - width(line) / 2f, y - height, color)
					is FormattedCharSequence -> draw(poseStack, line, x - width(line) / 2f, y - height, color)
				}

				y += lineHeight
			}
		}

		TextAlign.BOTTOM_RIGHT -> {
			for (line in text) {
				when (line) {
					is Component -> draw(poseStack, line, x - width(line), y - height, color)
					is String -> draw(poseStack, line, x - width(line), y - height, color)
					is FormattedCharSequence -> draw(poseStack, line, x - width(line), y - height, color)
				}

				y += lineHeight
			}
		}
	}
}

fun Font.drawAlignedStripe(poseStack: PoseStack, text: List<Any>, align: TextAlign, x: Float, y: Float, color: Int) {
	var totalWidth = 0

	for (line in text) {
		totalWidth += when (line) {
			is Component -> {
				width(line)
			}

			is String -> {
				width(line)
			}

			is FormattedCharSequence -> {
				width(line)
			}

			else -> throw IllegalArgumentException("Invalid stripe value ${line}")
		}
	}

	if (totalWidth == 0) {
		return
	}

	var x = x

	when (align) {
		TextAlign.TOP_LEFT -> {
			for (line in text) {
				when (line) {
					is Component -> {
						draw(poseStack, line, x, y, color)
						x += width(line)
					}
					is String -> {
						draw(poseStack, line, x, y, color)
						x += width(line)
					}
					is FormattedCharSequence -> {
						draw(poseStack, line, x, y, color)
						x += width(line)
					}
				}
			}
		}

		TextAlign.TOP_CENTER -> {
			x += totalWidth / 2f

			for (line in text) {
				when (line) {
					is Component -> {
						draw(poseStack, line, x, y, color)
						x -= width(line)
					}
					is String -> {
						draw(poseStack, line, x, y, color)
						x -= width(line)
					}
					is FormattedCharSequence -> {
						draw(poseStack, line, x, y, color)
						x -= width(line)
					}
				}
			}
		}

		TextAlign.TOP_RIGHT -> {
			x += totalWidth

			for (line in text) {
				when (line) {
					is Component -> {
						draw(poseStack, line, x, y, color)
						x -= width(line)
					}
					is String -> {
						draw(poseStack, line, x, y, color)
						x -= width(line)
					}
					is FormattedCharSequence -> {
						draw(poseStack, line, x, y, color)
						x -= width(line)
					}
				}
			}
		}

		TextAlign.CENTER_LEFT -> {
			for (line in text) {
				when (line) {
					is Component -> {
						draw(poseStack, line, x, y - lineHeight / 2f, color)
						x += width(line)
					}
					is String -> {
						draw(poseStack, line, x, y - lineHeight / 2f, color)
						x += width(line)
					}
					is FormattedCharSequence -> {
						draw(poseStack, line, x, y - lineHeight / 2f, color)
						x += width(line)
					}
				}
			}
		}

		TextAlign.CENTER_CENTER -> {
			x += totalWidth / 2f

			for (line in text) {
				when (line) {
					is Component -> {
						draw(poseStack, line, x, y - lineHeight / 2f, color)
						x -= width(line)
					}
					is String -> {
						draw(poseStack, line, x, y - lineHeight / 2f, color)
						x -= width(line)
					}
					is FormattedCharSequence -> {
						draw(poseStack, line, x, y - lineHeight / 2f, color)
						x -= width(line)
					}
				}
			}
		}

		TextAlign.CENTER_RIGHT -> {
			x += totalWidth

			for (line in text) {
				when (line) {
					is Component -> {
						draw(poseStack, line, x, y - lineHeight / 2f, color)
						x -= width(line)
					}
					is String -> {
						draw(poseStack, line, x, y - lineHeight / 2f, color)
						x -= width(line)
					}
					is FormattedCharSequence -> {
						draw(poseStack, line, x, y - lineHeight / 2f, color)
						x -= width(line)
					}
				}
			}
		}

		TextAlign.BOTTOM_LEFT -> {
			for (line in text) {
				when (line) {
					is Component -> {
						draw(poseStack, line, x, y - lineHeight, color)
						x += width(line)
					}
					is String -> {
						draw(poseStack, line, x, y - lineHeight, color)
						x += width(line)
					}
					is FormattedCharSequence -> {
						draw(poseStack, line, x, y - lineHeight, color)
						x += width(line)
					}
				}
			}
		}

		TextAlign.BOTTOM_CENTER -> {
			x += totalWidth / 2f

			for (line in text) {
				when (line) {
					is Component -> {
						draw(poseStack, line, x, y - lineHeight, color)
						x -= width(line)
					}
					is String -> {
						draw(poseStack, line, x, y - lineHeight, color)
						x -= width(line)
					}
					is FormattedCharSequence -> {
						draw(poseStack, line, x, y - lineHeight, color)
						x -= width(line)
					}
				}
			}
		}

		TextAlign.BOTTOM_RIGHT -> {
			x += totalWidth

			for (line in text) {
				when (line) {
					is Component -> {
						draw(poseStack, line, x, y - lineHeight, color)
						x -= width(line)
					}
					is String -> {
						draw(poseStack, line, x, y - lineHeight, color)
						x -= width(line)
					}
					is FormattedCharSequence -> {
						draw(poseStack, line, x, y - lineHeight, color)
						x -= width(line)
					}
				}
			}
		}
	}
}
