package ru.dbotthepony.mc.otm.core

import com.mojang.blaze3d.systems.RenderSystem
import net.minecraft.ChatFormatting

data class RGBAColor(val red: Float, val green: Float, val blue: Float, val alpha: Float = 1f) {
	constructor(r: Int, g: Int, b: Int) : this((r / 255f), (g / 255f), (b / 255f), 1f)
	constructor(r: Int, g: Int, b: Int, a: Int) : this((r / 255f), (g / 255f), (b / 255f), (a / 255f))
	constructor(r: Int, g: Int, b: Int, a: Float) : this((r / 255f), (g / 255f), (b / 255f), a)

	constructor(color: Long) : this(
		(((color and -0x1000000) ushr 24) / 255f),
		(((color and 0xFF0000) ushr 16) / 255f),
		(((color and 0xFF00) ushr 8) / 255f),
		(((color and 0xFF)) / 255f)
	)

	fun toInt(): Int {
		val r = (this.red * 255).toInt()
		val g = (this.green * 255).toInt()
		val b = (this.blue * 255).toInt()
		return (r shl 16) or (g shl 8) or b
	}

	fun toIntInv(): Int {
		val r = (this.red * 255).toInt()
		val g = (this.green * 255).toInt()
		val b = (this.blue * 255).toInt()
		return (b shl 16) or (g shl 8) or r
	}

	fun setSystemColor() {
		setShaderColor()
		setDrawColor()
	}

	fun setShaderColor() {
		RenderSystem.setShaderColor(red, green, blue, alpha)
	}

	fun setDrawColor() {
		ru.dbotthepony.mc.otm.client.render.setDrawColor(this)
	}

	fun linearInterpolation(t: Float, other: RGBAColor): RGBAColor {
		return RGBAColor(
			linearInterpolation(t, red, other.red),
			linearInterpolation(t, green, other.green),
			linearInterpolation(t, blue, other.blue),
			linearInterpolation(t, alpha, other.alpha),
		)
	}

	companion object {
		val BLACK = RGBAColor(0f, 0f, 0f, 1f)
		val WHITE = RGBAColor(1f, 1f, 1f, 1f)
		val RED = RGBAColor(1f, 0f, 0f)
		val GREEN = RGBAColor(0f, 1f, 0f, 1f)
		val LIGHT_GREEN = RGBAColor(136, 255, 124)
		val SLATE_GRAY = RGBAColor(64, 64, 64)
		val GRAY = RGBAColor(0x2C2C2CFFL)

		val DARK_BLUE = rgb(ChatFormatting.DARK_BLUE.color!!)
		val DARK_GREEN = rgb(ChatFormatting.DARK_GREEN.color!!)
		val DARK_AQUA = rgb(ChatFormatting.DARK_AQUA.color!!)
		val DARK_RED = rgb(ChatFormatting.DARK_RED.color!!)
		val DARK_PURPLE = rgb(ChatFormatting.DARK_PURPLE.color!!)
		val GOLD = rgb(ChatFormatting.GOLD.color!!)
		val DARK_GRAY = rgb(ChatFormatting.DARK_GRAY.color!!)
		val BLUE = rgb(ChatFormatting.BLUE.color!!)
		val AQUA = rgb(ChatFormatting.AQUA.color!!)
		val LIGHT_PURPLE = rgb(ChatFormatting.LIGHT_PURPLE.color!!)
		val YELLOW = rgb(ChatFormatting.YELLOW.color!!)

		val LOW_POWER = RGBAColor(173, 41, 41)
		val FULL_POWER = RGBAColor(255, 242, 40)
		val LOW_MATTER = RGBAColor(0, 24, 148)
		val FULL_MATTER = RGBAColor(72, 90, 255)
		val LOW_PATTERNS = RGBAColor(44, 104, 57)
		val FULL_PATTERNS = RGBAColor(65, 255, 87)

		fun inv(color: Int): RGBAColor {
			val r = (color and -0x1000000 ushr 24) / 255f
			val g = (color and 0xFF0000 ushr 16) / 255f
			val b = (color and 0xFF00 ushr 8) / 255f
			val a = (color and 0xFF) / 255f
			return RGBAColor(a, b, g, r)
		}

		fun rgb(color: Int): RGBAColor {
			val r = (color and 0xFF0000 ushr 16) / 255f
			val g = (color and 0xFF00 ushr 8) / 255f
			val b = (color and 0xFF) / 255f
			return RGBAColor(r, g, b)
		}
	}
}

fun linearInterpolation(t: Float, a: RGBAColor, b: RGBAColor): RGBAColor {
	return a.linearInterpolation(t, b)
}
